import 'dart:io';
import 'package:riedev_graph/graph.dart';

typedef String NodeLabelFunc<TNode>(TNode node, int idx);
typedef String EdgeLabelFunc<TEdge>(TEdge edge, int sourceIdx, int targetIdx);

enum NodeShape { diamond, box, circle, record, plaintext }

/// Define a style of node like color, shape and label.
class NodeStyle<TNode, TEdge extends Edge<TNode>> {
  final DotFile<TNode, TEdge> _exporter;

  NodeShape _shape;
  String _color;
  String _fillcolor;
  NodeLabelFunc<TNode> _labelFunc;

  NodeStyle(this._exporter) : _shape = NodeShape.circle;

  /// Set the shape of node.
  NodeStyle<TNode, TEdge> setShape(NodeShape shape) {
    _shape = shape;
    return this;
  }

  /// Set the border color of node.
  NodeStyle<TNode, TEdge> setColor(String color) {
    _color = color;
    return this;
  }

  /// Set the filling color of node.
  NodeStyle<TNode, TEdge> setFillColor(String color) {
    _fillcolor = color;
    return this;
  }

  /// Set a function that creates out of node and their index a label.
  NodeStyle<TNode, TEdge> setLabel(NodeLabelFunc<TNode> labelFunc) {
    _labelFunc = labelFunc;
    return this;
  }

  /// This method represent the back navigation to the [DotFile], so that you can specify another settings for export.
  DotFile<TNode, TEdge> apply() => _exporter;
}

enum EdgeShape { dashed, filled }

/// Define a style of edge like color, shape and label.
class EdgeStyle<TNode, TEdge extends Edge<TNode>> {
  final DotFile<TNode, TEdge> _exporter;

  EdgeShape _shape;
  String _color;
  EdgeLabelFunc<TEdge> _labelFunc;
  bool _nonArrowHead;
  double _penWidth;
  bool _isPenWidthSet;

  EdgeStyle(this._exporter)
      : _shape = EdgeShape.filled,
        _isPenWidthSet = false,
        _nonArrowHead = false;

  /// Set the shape of edge.
  EdgeStyle<TNode, TEdge> setShape(EdgeShape shape) {
    _shape = shape;
    return this;
  }

  /// Set the color of edge.
  EdgeStyle<TNode, TEdge> setColor(String color) {
    _color = color;
    return this;
  }

  /// Set a function that creates out of edge and the source and target index a label.
  EdgeStyle<TNode, TEdge> setLabel(EdgeLabelFunc<TEdge> labelFunc) {
    _labelFunc = labelFunc;
    return this;
  }

  /// Set that the edge has no arrows.
  EdgeStyle<TNode, TEdge> nonArrowHead() {
    _nonArrowHead = true;
    return this;
  }

  /// Set the width of drawing pen (arrow width).
  EdgeStyle<TNode, TEdge> penWidth(double width) {
    _penWidth = width;
    _isPenWidthSet = true;
    return this;
  }

  /// This method represent the back navigation to the [DotFile], so that you can specify another settings for export.
  DotFile<TNode, TEdge> apply() => _exporter;
}

/// This class is a helper class for building a string representation of the dot file node style.
class NodeStyleBuilder<TNode, TEdge extends Edge<TNode>>
    extends NodeStyle<TNode, TEdge> {
  NodeStyleBuilder(DotFile<TNode, TEdge> exporter) : super(exporter);

  /// Build a string out of the [NodeStyle].
  String Build(TNode node, int idx) {
    return "$idx " + _getStyleString(node, idx);
  }

  String _getStyleString(TNode node, int idx) {
    var style = "[";

    if (_labelFunc != null) {
      style += "label=\"${_labelFunc(node, idx)}\"";
    }
    if (_color != null) {
      style += style.length == 1 ? "color=\"$_color\"" : " color=\"$_color\"";
    }
    if (_fillcolor != null) {
      style += style.length == 1
          ? "style=\"filled\" fillcolor=\"$_fillcolor\""
          : " style=\"filled\" fillcolor=\"$_fillcolor\"";
    }

    style += style.length == 1
        ? "shape=\"${_getEnumValue(_shape)}\""
        : " shape=\"${_getEnumValue(_shape)}\"";

    return style + "];";
  }

  String _getEnumValue(NodeShape shape) {
    return "${shape.toString().substring(shape.toString().indexOf('.') + 1)}";
  }
}

/// This class is a helper class for building a string representation of the dot file edge style.
class EdgeStyleBuilder<TNode, TEdge extends Edge<TNode>>
    extends EdgeStyle<TNode, TEdge> {
  EdgeStyleBuilder(DotFile<TNode, TEdge> exporter) : super(exporter);

  /// Build a string out of the [EdgeStyle].
  String Build(TEdge edge, int sourceIdx, int targetIdx) {
    return "$sourceIdx -> $targetIdx " +
        _getStyleString(edge, sourceIdx, targetIdx);
  }

  String _getStyleString(TEdge edge, int sourceIdx, int targetIdx) {
    var style = "[";

    if (_labelFunc != null) {
      style += "label=\"${_labelFunc(edge, sourceIdx, targetIdx)}\"";
    }
    if (_color != null) {
      style += style.length == 1 ? "color=\"$_color\"" : " color=\"$_color\"";
    }

    style += style.length == 1
        ? "style=\"${_getEnumValue(_shape)}\""
        : " style=\"${_getEnumValue(_shape)}\"";

    if (_nonArrowHead) {
      style += style.length == 1 ? "arrowhead=\"none\"" : " arrowhead=\"none\"";
    }

    if (_isPenWidthSet) {
      style += style.length == 1
          ? "penwidth=\"$_penWidth\""
          : " penwidth=\"$_penWidth\"";
    }

    return style + "];";
  }

  String _getEnumValue(EdgeShape shape) {
    return "${shape.toString().substring(shape.toString().indexOf('.') + 1)}";
  }
}

class DotFile<TNode, TEdge extends Edge<TNode>> {
  String _filename;

  NodeStyleBuilder<TNode, TEdge> _nodeStyle;
  final Map<TNode, NodeStyleBuilder<TNode, TEdge>> _nodeStyles = Map();

  EdgeStyleBuilder<TNode, TEdge> _edgeStyle;
  final Map<TEdge, EdgeStyleBuilder<TNode, TEdge>> _edgeStyles = Map();

  final Map<Graph<TNode, TEdge>, String> _graphs = Map();

  DotFile() {
    _nodeStyle = NodeStyleBuilder<TNode, TEdge>(this);
    _edgeStyle = EdgeStyleBuilder<TNode, TEdge>(this);
  }

  /// Export graph to file.
  void export() {
    var file = File(_filename ?? 'graph.dot');
    var sink = file.openWrite();

    for (var graph in _graphs.keys) {
      _writeGraph(sink, graph, _graphs[graph]);
      sink.writeln();
    }

    sink.close();
  }

  /// Add a graph to the dot file.
  DotFile<TNode, TEdge> addGraph(Graph<TNode, TEdge> graph,
      {String graphName = "g"}) {
    _graphs[graph] = graphName;
    return this;
  }

  /// Set the name of the file.
  DotFile<TNode, TEdge> setFileName(String filename) {
    _filename = filename;
    return this;
  }

  /// Set node style either for all nodes or for a specific [node].
  NodeStyle<TNode, TEdge> setNodeStyle({TNode node}) {
    var style = NodeStyleBuilder<TNode, TEdge>(this);

    if (node == null) {
      _nodeStyle = style;
    } else {
      _nodeStyles[node] = style;
    }

    return style;
  }

  /// Set edge style either for all edges or for a specific [edge].
  EdgeStyle<TNode, TEdge> setEdgeStyle({TEdge edge}) {
    var style = EdgeStyleBuilder<TNode, TEdge>(this);

    if (edge == null) {
      _edgeStyle = style;
    } else {
      _edgeStyles[edge] = style;
    }

    return style;
  }

  void _writeGraph(IOSink sink, Graph<TNode, TEdge> graph, String graphName) {
    sink.writeln("digraph $graphName {");

    var idx = 0;
    var nodeIndices = Map<TNode, int>();

    for (var node in graph.Nodes) {
      sink.writeln(_getNodeString(node, idx));
      nodeIndices[node] = idx++;
    }

    for (var edge in graph.Edges) {
      sink.writeln(_getEdgeString(
          edge, nodeIndices[edge.source], nodeIndices[edge.target]));
    }

    sink.writeln("}");
  }

  String _getNodeString(TNode node, int idx) {
    if (_nodeStyles.containsKey(node)) {
      return _nodeStyles[node].Build(node, idx);
    }
    return _nodeStyle.Build(node, idx);
  }

  String _getEdgeString(TEdge edge, int sourceIdx, int targetIdx) {
    if (_edgeStyles.containsKey(edge)) {
      return _edgeStyles[edge].Build(edge, sourceIdx, targetIdx);
    }
    return _edgeStyle.Build(edge, sourceIdx, targetIdx);
  }
}
