/// A edge from source to target node.
class Edge<TNode> {
  final TNode _source;
  final TNode _target;

  const Edge(this._source, this._target);

  TNode get source => _source;
  TNode get target => _target;
}


/// For every node the graph will save the first in edge and the first out edge reference. 
/// Futhermore the graph save for every edge a list of two item the next out edge and the next in edge reference.
/// This creates a linked list of out and in edges for every node of the graph.
class Graph<TNode, TEdge extends Edge<TNode>> {
  final bool _allowDoubleEdges; 

  final Map<TNode, List<TEdge>> _linkedNodes = Map();
  final Map<TEdge, List<TEdge>> _linkedEdges = Map();

  Graph(this._allowDoubleEdges,
      {Iterable<TNode> nodes, Iterable<TEdge> edges}) {
    if (nodes != null) {
      addNodes(nodes);
    }
    if (edges != null) {
      addEdges(edges);
    }
  }

  Iterable<TNode> get Nodes => _linkedNodes.keys;
  Iterable<TEdge> get Edges => _linkedEdges.keys;

  /// Add the [node] to the graph if the [node] is not added yet.
  void addNode(TNode node) {
    if (node == null) {
      return;
    }
    if (_linkedNodes.containsKey(node)) {
      return;
    }
    _linkedNodes[node] = List<TEdge>(2);
  }

  /// Add the [edge] to the graph if the [edge] is not added yet. Futhermore the method check if a edge is exists from source to target node and add these new edge only if its allowed.
  void addEdge(TEdge edge) {
    if (edge == null) {
      return;
    }

    if (_linkedEdges.containsKey(edge)) {
      return;
    }

    if (!_allowDoubleEdges && existEdge(edge.source, edge.target)) {
      return;
    }

    addNode(edge.source);
    addNode(edge.target);

    _linkedEdges[edge] = List<TEdge>(2);

    _insertEdge(edge, true); //Add InEdge reference
    _insertEdge(edge, false); //Add OutEdge reference
  }

  /// Remove the [node] of the graph.
  void removeNode(TNode node) {
    if (node == null) {
      return;
    }

    if (!_linkedNodes.containsKey(node)) {
      return;
    }

    for (var edge in inEdges(node)) {
      removeEdge(edge);
    }

    for (var edge in outEdges(node)) {
      removeEdge(edge);
    }

    _linkedNodes.remove(node);
  }

  /// Remove the [edge] of the graph.
  void removeEdge(TEdge edge) {
    if (edge == null) {
      return;
    }

    if (!_linkedEdges.containsKey(edge)) {
      return;
    }

    _removeEdge(edge, true);
    _removeEdge(edge, false);

    _linkedEdges.remove(edge);
  }

  /// Add the next in or out edge reference of the corresponding node.
  /// [isIn] = false => add next out edge reference
  /// [isIn] = true => add next in edge reference
  void _insertEdge(TEdge edge, bool isIn) {
    //0: Out 1: In
    var index = isIn ? 1 : 0;

    //Get the the first (head) in and out edge of the node.
    var edgeRefs = _linkedNodes[isIn ? edge.target : edge.source];

    //Get first in/out edge (out: index == 0, in: index == 1).
    var head = edgeRefs[index];

    //Has the node any in/out edge?
    if (head == null) {
      //No, the node has no in/out edges yet.
      edgeRefs[index] = edge; //Set the first in/out edge.
    } else {
      //Yes, the node already any in/out edges.
      while (true) {
        //Step throw the linked list and find the next free place for the in/out edge.
        var next = _linkedEdges[head][index];
        if (next == null) {
          break;
        }

        head = next;
      }
      _linkedEdges[head][index] = edge; //Set the next in/out edge.
    }
  }

  /// Remove the next in or out edge reference of the corresponding node and replace it with the over next reference.
  /// [isIn] = false => remove next out edge reference
  /// [isIn] = true => remove next in edge reference
  void _removeEdge(TEdge edge, bool isIn) {
    //0: Out 1: In
    var index = isIn ? 1 : 0;

    //Get the the first (head) in and out edge of the node.
    var edgeRefs = _linkedNodes[
        isIn ? edge.target : edge.source]; //_linkedNodes[Corresponding node]

    //Get first in/out edge (out: index == 0, in: index == 1).
    var head = edgeRefs[index];

    //Is the head edge the wanted in/out edge?
    if (identical(head, edge)) {
      //Yes, replace the edge reference with the over next reference.
      edgeRefs[index] = _linkedEdges[edge][index]; //replace.
    } else {
      //No, find the wanted edge reference and replace it with the over next reference.
      while (true) {
        //Step throw the linked list and find the wanted in/out edge.
        var next = _linkedEdges[head][index];
        if (identical(next, edge)) {
          _linkedEdges[head][index] = _linkedEdges[edge][index]; //replace
          break;
        }

        head = next;
      }
    }
  }

  /// Add all [nodes] to the graph.
  void addNodes(Iterable<TNode> nodes) {
    for (var node in nodes) {
      addNode(node);
    }
  }

  /// Add all [edges] to the graph.
  void addEdges(Iterable<TEdge> edges) {
    for (var edge in edges) {
      addEdge(edge);
    }
  }

  /// Return the number of edges that leading in the [node].
  int inDegree(TNode node) {
    if (node == null) {
      return 0;
    }
    if (!_linkedNodes.containsKey(node)) {
      return 0;
    }

    var count = 0;
    var head = _linkedNodes[node][1];

    while (head != null) {
      count++;
      head = _linkedEdges[head][1];
    }

    return count;
  }

  /// Return the number of edges that leading out the [node].
  int outDegree(TNode node) {
    if (node == null) {
      return 0;
    }
    if (!_linkedNodes.containsKey(node)) {
      return 0;
    }

    var count = 0;
    var head = _linkedNodes[node][0];

    while (head != null) {
      count++;
      head = _linkedEdges[head][0];
    }

    return count;
  }

  /// Return all edges that leading out the [node] ([node] == edge.Source).
  Iterable<TEdge> outEdges(TNode node) {
    if (node == null) {
      return [];
    }
    if (!_linkedNodes.containsKey(node)) {
      return [];
    }

    var edges = <TEdge>[];
    var head = _linkedNodes[node][0];

    while (head != null) {
      edges.add(head);
      head = _linkedEdges[head][0];
    }

    return edges;
  }

  /// Return all edges that leading in the [node] ([node] == edge.Target).
  Iterable<TEdge> inEdges(TNode node) {
    if (node == null) {
      return [];
    }
    if (!_linkedNodes.containsKey(node)) {
      return [];
    }

    var edges = <TEdge>[];
    var head = _linkedNodes[node][1];

    while (head != null) {
      edges.add(head);
      head = _linkedEdges[head][1];
    }

    return edges;
  }

  /// Return all subsequent nodes of [node] if there exist else a empty list will return.
  Iterable<TNode> subsequentNodes(TNode node) =>
      outEdges(node).map((x) => x.target);

  /// Return all predecessor nodes of [node] if there exist else a empty list will return.
  Iterable<TNode> predeccesorNodes(TNode node) =>
      inEdges(node).map((x) => x.source);

  /// Return all nodes that no edges coming in. 
  Iterable<TNode> getSources() =>
      _linkedNodes.keys.where((x) => inDegree(x) == 0);

  /// Return all nodes that no edges leading out. 
  Iterable<TNode> getTails() =>
      _linkedNodes.keys.where((x) => outDegree(x) == 0);

  /// Check if a edge from [source] to [target] already exist inside the graph.
  bool existEdge(TNode source, TNode target) =>
      subsequentNodes(source).contains(target);
}
