import 'package:riedev_graph/graph.dart';
import "package:test/test.dart";

/* 
All graphs are directed graphs and their edges are defined from up to down or from left to right.
*/
void main() {
/*
1 -- 2
*/
  group('simple graph test', () {
    test('allow double edges test', () {
      var g = Graph<int, Edge<int>>(true); //<-- allow double edges
      g.addEdge(Edge(1, 2));

      assert(g.Nodes.length == 2);

      assert(g.Edges.length == 1);
      g.addEdge(Edge(1, 2));
      assert(g.Edges.length == 2);
      g.addEdge(Edge(1, 2));
      assert(g.Edges.length == 3);
    });

    test('not allow double edges test', () {
      var g = Graph<int, Edge<int>>(false); //<-- not allow double edges
      g.addEdge(Edge(1, 2));

      assert(g.Nodes.length == 2);

      assert(g.Edges.length == 1);
      g.addEdge(Edge(1, 2));
      assert(g.Edges.length == 1);
      g.addEdge(Edge(1, 2));
      assert(g.Edges.length == 1);
    });
  });

/*
        1
      /   \ 
     2     3
    / \   / \
   4   5 6   7
*/
  group('binary graph test: ', () {
    Edge<int> e1;
    Edge<int> e2;
    Edge<int> e3;
    Edge<int> e4;
    Edge<int> e5;
    Edge<int> e6;

    Graph<int, Edge<int>> g;

    setUp(() {
      g = Graph<int, Edge<int>>(false);

      e1 = const Edge(1, 2);
      e2 = const Edge(1, 3);
      e3 = const Edge(2, 4);
      e4 = const Edge(2, 5);
      e5 = const Edge(3, 6);
      e6 = const Edge(3, 7);

      g.addEdge(e1);
      g.addEdge(e2);
      g.addEdge(e3);
      g.addEdge(e4);
      g.addEdge(e5);
      g.addEdge(e6);
    });

    test('contains all nodes', () {
      assert(g.Nodes.length == 7);

      assert(g.Nodes.contains(1));
      assert(g.Nodes.contains(2));
      assert(g.Nodes.contains(3));
      assert(g.Nodes.contains(4));
      assert(g.Nodes.contains(5));
      assert(g.Nodes.contains(6));
      assert(g.Nodes.contains(7));
    });

    test('contains all edges', () {
      assert(g.Edges.length == 6);

      assert(g.Edges.contains(e1));
      assert(g.Edges.contains(e2));
      assert(g.Edges.contains(e3));
      assert(g.Edges.contains(e4));
      assert(g.Edges.contains(e5));
      assert(g.Edges.contains(e6));
    });

    test('getSources test', () {
      var sources = g.getSources();

      assert(sources.length == 1);
      assert(sources.first == 1);
    });

    test('getTails test', () {
      var tails = g.getTails();

      assert(tails.length == 4);

      assert(tails.contains(4));
      assert(tails.contains(5));
      assert(tails.contains(6));
      assert(tails.contains(7));
    });

    test('predeccesorNodes test', () {
      var nodes = g.predeccesorNodes(2);

      assert(nodes.length == 1);
      assert(nodes.first == 1);
    });

    test('subsequentNodes test', () {
      var nodes = g.subsequentNodes(2);

      assert(nodes.length == 2);
      assert(nodes.contains(4));
      assert(nodes.contains(5));
    });

    test('inEdges test', () {
      var edges = g.inEdges(2);

      assert(edges.length == 1);
      assert(edges.contains(e1));
    });

    test('outEdges test', () {
      var edges = g.outEdges(2);

      assert(edges.length == 2);
      assert(edges.contains(e3));
      assert(edges.contains(e4));
    });

    test('inDegree test', () {
      var degree = g.inDegree(2);

      assert(degree == 1);
    });

    test('outDegree test', () {
      var degree = g.outDegree(2);

      assert(degree == 2);
    });

    test('double add exist node test', () {
      assert(g.Nodes.length == 7);
      g.addNode(1);
      assert(g.Nodes.length == 7);
    });

    test('double add exist edge test', () {
      assert(g.Edges.length == 6);
      g.addEdge(e1);
      assert(g.Edges.length == 6);
    });

    test('double add edge test', () {
      assert(g.Edges.length == 6);
      g.addEdge(Edge(1, 2));
      assert(g.Edges.length == 6);
    });

    test('existEdge test', () {
      assert(g.existEdge(1, 2));
    });
  });

/*
        1
      /   \ 
     2     3  10
    / \   / \ /
   4    5    6
   |    |    |
   11   7    |
   |    |    |
   |    8    |
   |    \   /
   |      9
   12    | \
    \    | 14
     \  /
      13 
*/
  group('complex graph test: ', () {
    Edge<int> e1;
    Edge<int> e2;
    Edge<int> e3;
    Edge<int> e4;
    Edge<int> e5;
    Edge<int> e6;
    Edge<int> e7;
    Edge<int> e8;
    Edge<int> e9;
    Edge<int> e10;
    Edge<int> e11;
    Edge<int> e12;
    Edge<int> e13;
    Edge<int> e14;
    Edge<int> e15;
    Edge<int> e16;

    List<Edge<int>> edges;

    Graph<int, Edge<int>> g;

    setUp(() {
      g = Graph<int, Edge<int>>(false);

      e1 = const Edge(1, 2);
      e2 = const Edge(1, 3);
      e3 = const Edge(2, 4);
      e4 = const Edge(2, 5);
      e5 = const Edge(3, 5);
      e6 = const Edge(3, 6);
      e7 = const Edge(10, 6);
      e8 = const Edge(4, 11);
      e9 = const Edge(5, 7);
      e10 = const Edge(6, 9);
      e11 = const Edge(11, 12);
      e12 = const Edge(7, 8);
      e13 = const Edge(8, 9);
      e14 = const Edge(9, 14);
      e15 = const Edge(12, 13);
      e16 = const Edge(9, 13);

      edges = [
        e1,
        e2,
        e3,
        e4,
        e5,
        e6,
        e7,
        e8,
        e9,
        e10,
        e11,
        e12,
        e13,
        e14,
        e15,
        e16
      ];
      g.addEdges(edges);
    });

    test('contains all nodes', () {
      assert(g.Nodes.length == 14);

      for (int i = 1; i <= 14; i++) {
        assert(g.Nodes.contains(i));
      }
    });

    test('contains all edges', () {
      assert(g.Edges.length == 16);

      for (var e in edges) {
        assert(g.Edges.contains(e));
      }
    });

    test('getSources test', () {
      var sources = g.getSources();

      assert(sources.length == 2);
      assert(sources.contains(1));
      assert(sources.contains(10));
    });

    test('getTails test', () {
      var tails = g.getTails();

      assert(tails.length == 2);

      assert(tails.contains(13));
      assert(tails.contains(14));
    });

    test('predeccesorNodes test', () {
      var nodes = g.predeccesorNodes(5);

      assert(nodes.length == 2);
      assert(nodes.contains(2));
      assert(nodes.contains(3));
    });

    test('subsequentNodes test', () {
      var nodes = g.subsequentNodes(9);

      assert(nodes.length == 2);
      assert(nodes.contains(13));
      assert(nodes.contains(14));
    });

    test('inEdges test', () {
      var edges = g.inEdges(13);

      assert(edges.length == 2);
      assert(edges.contains(e15));
      assert(edges.contains(e16));
    });

    test('outEdges test', () {
      var edges = g.outEdges(9);

      assert(edges.length == 2);
      assert(edges.contains(e14));
      assert(edges.contains(e16));
    });

    test('inDegree test', () {
      var degree = g.inDegree(6);
      assert(degree == 2);
    });

    test('outDegree test', () {
      var degree = g.outDegree(2);
      assert(degree == 2);
    });

    test('double add exist node test', () {
      assert(g.Nodes.length == 14);
      g.addNode(1);
      assert(g.Nodes.length == 14);
    });

    test('double add exist edge test', () {
      assert(g.Edges.length == 16);
      g.addEdge(e1);
      assert(g.Edges.length == 16);
    });

    test('double add edge test', () {
      assert(g.Edges.length == 16);
      g.addEdge(Edge(1, 2));
      assert(g.Edges.length == 16);
    });

    test('existEdge test', () {
      assert(g.existEdge(1, 2));
    });
    
    test('remove edge test', () {
      assert(g.Edges.length == 16);
      assert(g.Nodes.length == 14);
      g.removeEdge(e16);
      assert(g.Edges.length == 15);
      assert(g.Nodes.length == 14);

      assert(g.Edges.contains(e1));
      assert(g.Edges.contains(e2));
      assert(g.Edges.contains(e3));
      assert(g.Edges.contains(e4));
      assert(g.Edges.contains(e5));
      assert(g.Edges.contains(e6));
      assert(g.Edges.contains(e7));
      assert(g.Edges.contains(e8));
      assert(g.Edges.contains(e9));
      assert(g.Edges.contains(e10));
      assert(g.Edges.contains(e11));
      assert(g.Edges.contains(e12));
      assert(g.Edges.contains(e13));
      assert(g.Edges.contains(e14));
      assert(g.Edges.contains(e15));

      assert(!g.Edges.contains(e16));
    });

    test('remove node test', () {
      assert(g.Edges.length == 16);
      assert(g.Nodes.length == 14);
      g.removeNode(5);
      assert(g.Edges.length == 13);
      assert(g.Nodes.length == 13);

      assert(g.Edges.contains(e1));
      assert(g.Edges.contains(e2));
      assert(g.Edges.contains(e3));
      assert(g.Edges.contains(e6));
      assert(g.Edges.contains(e7));
      assert(g.Edges.contains(e8));
      assert(g.Edges.contains(e10));
      assert(g.Edges.contains(e11));
      assert(g.Edges.contains(e12));
      assert(g.Edges.contains(e13));
      assert(g.Edges.contains(e14));
      assert(g.Edges.contains(e15));
      assert(g.Edges.contains(e16));

      assert(!g.Edges.contains(e4));
      assert(!g.Edges.contains(e5));
      assert(!g.Edges.contains(e9));

      assert(g.inDegree(7) == 0);
      assert(g.outDegree(2) == 1);
      assert(g.outDegree(3) == 1);
    });
  });
}
